/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.ModifiedBy;
import dk.sds.ecpr2.dao.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractCrossTableDAOImpl;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.*;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Repository
@Log4j2
public class SearchCrossTableDAOImpl extends AbstractCrossTableDAOImpl {

    @Autowired
    PersonDAO personDAO;

    @Autowired
    private IdentifierDAO identifierDAO;

    @Autowired
    private AddressDAO addressDAO;

    @Autowired
    private BirthDateDAO birthDateDAO;

    @Autowired
    private ContactDAO contactDAO;

    @Autowired
    private GenderDAO genderDAO;

    @Autowired
    private NameDAO nameDAO;

    @Autowired
    protected ModifierDAO modifierDAO;


    @Override
    public List<PersonVO> getPersonsWithRelationsAfterPID(int minPersonPid,
                                                          int batchSize,
                                                          boolean excludeReservedUntilFilter,
                                                          boolean excludeDuplicatePersonsFilter) throws DAOException {
        return getPersonsWithRelations(minPersonPid, batchSize, excludeReservedUntilFilter, excludeDuplicatePersonsFilter);
    }

    @Override
    public List<PersonVO> getPersonsWithRelationsAfterModifiedTime(Instant modifiedDateTime, boolean excludeDuplicatePersonsFilter, boolean excludeReservedUntil) throws DAOException {
        return getPersonsWithRelations(modifiedDateTime, excludeDuplicatePersonsFilter, excludeReservedUntil);
    }

    private List<PersonVO> getPersonsWithRelations(int minPersonPid,
                                                   int batchSize,
                                                   boolean excludeReservedUntil,
                                                   boolean excludeDuplicatePersonsFilter) throws DAOException {
        // Get all persons after PID
        List<PersonVO> personList = personDAO.getPersonsAfterPID(minPersonPid, batchSize, excludeReservedUntil);

        final int maxPersonPID = PersonVO.getMaxPersonPID(personList);
        final int minPersonPID = PersonVO.getMinPersonPID(personList);

        // Enrich all those persons with their data
        List<List<PersonVO>> personPersonList = personList.stream().map(PersonVO::getPersonUUID).distinct().map(uuid -> {
            try {
                return getPersonWithRelations(uuid);
            } catch (DAOException e) {
                throw new RuntimeException(e);
            }
        }).collect(Collectors.toList());

        // Merge each group of persons to an actual person
        personList = personPersonList.stream().map(pl -> {
            try {
                return mergePersons(pl, maxPersonPID);
            } catch (DAOException e) {
                throw new RuntimeException(e);
            }
        }).collect(Collectors.toList());
        return personList;
    }

    private List<PersonVO> getPersonsWithRelations(Instant modifiedTime, boolean excludeDuplicatePersons, boolean excludeReservedUntil) throws DAOException {
        // Get all persons modified after modifiedTime
        List<PersonVO> personList = personDAO.getPersonsModifiedAfter(modifiedTime, excludeReservedUntil);

        final int maxPersonPID = PersonVO.getMaxPersonPID(personList);

        // Enrich all those persons with their data
        List<List<PersonVO>> personPersonList = personList.stream().map(PersonVO::getPersonUUID).distinct().map(uuid -> {
            try {
                return getPersonWithRelations(uuid);
            } catch (DAOException e) {
                throw new RuntimeException(e);
            }
        }).collect(Collectors.toList());

        // Merge each group of persons to an actual person
        personList = personPersonList.stream().map(pl -> {
            try {
                return mergePersons(pl, maxPersonPID);
            } catch (DAOException e) {
                throw new RuntimeException(e);
            }
        }).collect(Collectors.toList());
        return personList;
    }

    // Merge Person rows into a single PersonVO containing all related entities
    public PersonVO mergePersons(List<PersonVO> personList, int maxPersonPID) throws DAOException {
        if (personList.isEmpty()) {
            return null;
        }
        PersonVO mergedPerson = new PersonVO();
        PersonVO first = personList.get(0);
        mergedPerson.setPersonUUID(first.getPersonUUID());
        mergedPerson.setReservedUntil(first.getReservedUntil());
        mergedPerson.setPersonPID(maxPersonPID);
        for (PersonVO dbPerson : personList) {
            mergedPerson.getGenders().addAll(dbPerson.getGenders());
            mergedPerson.getIdentifiers().addAll(dbPerson.getIdentifiers());
            mergedPerson.getBirthDates().addAll(dbPerson.getBirthDates());
            mergedPerson.getAddresses().addAll(dbPerson.getAddresses());
            mergedPerson.getContacts().addAll(dbPerson.getContacts());
            mergedPerson.getNames().addAll(dbPerson.getNames());
        }
        addModifiers(Collections.singletonList(mergedPerson), false);
        return mergedPerson;
    }

    protected void addModifiers(List<PersonVO> persons, boolean includePersonIdentifier) throws DAOException {
        for (PersonVO person : persons) {
            List<Modified> versions = new ArrayList<>();
            List<PersonVO> personsByUUID = personDAO.getPersonsByUUIDs(Collections.singletonList(person.getPersonUUID()));
            for (int i = 0; i < personsByUUID.size(); i++) {
                PersonVO personVersion = personsByUUID.get(i);
                Modified modified = new Modified();
                ModifiedBy modifiedBy = modifierDAO.getModifier(personVersion.getModifierPID(), includePersonIdentifier);
                modified.setBy(modifiedBy);
                Calendar cal = new GregorianCalendar();
                cal.setTime(personVersion.getModifiedTime());
                modified.setValidFrom(cal);
                if (personsByUUID.size() == (i + 1)) {
                    // last modified, set ValidTo=INF
                    cal = new GregorianCalendar();
                    cal.setTime(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
                    modified.setValidTo(cal);
                } else {
                    cal = new GregorianCalendar();
                    cal.setTime(personsByUUID.get(i + 1).getModifiedTime());
                    modified.setValidTo(cal);
                }
                if (personVersion.getReservedUntil() != null) {
                    cal = new GregorianCalendar();
                    cal.setTime(personVersion.getReservedUntil());
                    modified.setReservedUntil(cal);
                }
                versions.add(modified);
            }
            person.getModifiers().addAll(versions);
        }
    }
}
