/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.security;

import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import dk.sdsd.dgws._2010._08.PredefinedRequestedRole;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.util.StringUtils;
import org.springframework.ws.context.MessageContext;
import org.springframework.ws.soap.*;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.annotation.PostConstruct;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import java.util.*;

@Log4j2
public class RoleValidationInterceptor implements org.springframework.ws.server.SmartEndpointInterceptor {

    private static final List<String> NON_INTERCEPTING_SOAP_ACTIONS = Collections.singletonList("http://sds.dk/ecpr2/2017/10/01#GetPermissions");

    private final Map<PredefinedRequestedRole, Authenticator> roleToAuthenticatorMap = new HashMap<>();

    @Autowired
    AdminAuthenticator adminAuthenticator;
    @Autowired
    DoctorAuthenticator doctorAuthenticator;
    @Autowired
    DentistAuthenticator dentistAuthenticator;
    @Autowired
    MidwifeAuthenticator midwifeAuthenticator;
    @Autowired
    NurseAuthenticator nurseAuthenticator;
    @Autowired
    ECprCoordinatorAuthenticator eCprCoordinatorAuthenticator;
    @Autowired
    SystemAuthenticator systemAuthenticator;
    @Autowired
    AssistantAuthenticator assistantAuthenticator;

    @PostConstruct
    public void postConstruct() {
        roleToAuthenticatorMap.put(PredefinedRequestedRole.WEB_ADMINISTRATOR, adminAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.LÆGE, doctorAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.TANDLÆGE, dentistAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.JORDEMODER, midwifeAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.SYGEPLEJERSKE, nurseAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.E_CPR_KOORDINATOR, eCprCoordinatorAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.SYSTEM, systemAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_JORDEMODER, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_LÆGE, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_SYGEPLEJERSKE, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_TANDLÆGE, assistantAuthenticator);
    }

    public Map<PredefinedRequestedRole, Authenticator> getRoleToAuthenticatorMap() {
        return roleToAuthenticatorMap;
    }

    @Override
    public boolean handleRequest(MessageContext messageContext, Object o) throws Exception {
        if (!shouldIntercept(messageContext, o)) {
            SoapMessage request = (SoapMessage) messageContext.getRequest();
            SoapHeader soapHeader = request.getSoapHeader();
            DOMSource soapHeaderSource = (DOMSource) soapHeader.getSource();
            NodeList requestedRoleList = soapHeaderSource.getNode().getOwnerDocument().getElementsByTagNameNS("http://www.sdsd.dk/dgws/2010/08", "RequestedRole");
            if (requestedRoleList == null || requestedRoleList.getLength() == 0) {
                log.warn("Der er ikke medgivet nogen RequestedRole");
                return returnError(messageContext, "Der er ikke medgivet noget RequestedRole");
            }
            String requestedRole = requestedRoleList.item(0).getTextContent();
            RequestContext.get().setRequestedRole(requestedRole);
            return true;
        }
        try {
            log.info("Autentikerer");
            SoapMessage request = (SoapMessage) messageContext.getRequest();
            SoapHeader soapHeader = request.getSoapHeader();
            DOMSource soapHeaderSource = (DOMSource) soapHeader.getSource();
            NodeList requestedRoleList = soapHeaderSource.getNode().getOwnerDocument().getElementsByTagNameNS("http://www.sdsd.dk/dgws/2010/08", "RequestedRole");
            if (requestedRoleList == null || requestedRoleList.getLength() == 0) {
                log.warn("Der er ikke medgivet nogen RequestedRole");
                return returnError(messageContext, "Der er ikke medgivet noget RequestedRole");
            }
            String requestedRole = requestedRoleList.item(0).getTextContent();

            List<String> errorMessages = new ArrayList<>();
            OnBehalfOf onBehalfOf = getOnBehalfOf(soapHeaderSource, errorMessages);
            SecurityContext securityContext = Security.getSecurityContext();
            Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
            if (!ticketOptional.isPresent()) {
                throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
            }
            SecurityContext.Ticket ticket = ticketOptional.get();
            if (!ticket.isValid()) {
                throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
            }
            if (errorMessages.size() > 0) {
                return returnError(messageContext, errorMessages.get(0));
            }
            PredefinedRequestedRole predefinedRequestedRole = null;
            try {
                predefinedRequestedRole = PredefinedRequestedRole.fromValue(requestedRole);
            } catch (IllegalArgumentException ignored) {
                // The below if statement handles unknown roles
            }

            Authenticator authenticator = roleToAuthenticatorMap.get(predefinedRequestedRole);
            if (authenticator == null) {
                log.warn("Rollen " + requestedRole + " kan ikke genkendes");
                // Return requestedRole, because that is what the client sent to the backend
                return returnError(messageContext, "Rollen " + requestedRole + " kan ikke genkendes");
            }
            log.info("Autentikerer for rollen " + requestedRole);
            if (!authenticator.authenticate(securityContext, requestedRole, onBehalfOf)) {
                if (securityContext.getActingUser().isPresent() && securityContext.getActingUser().get().getIdentifier() != null) {
                    log.warn(securityContext.getActingUser().get().getIdentifier() + " er ikke en " + requestedRole);
                    // Return requestedRole, because that is what the client sent to the backend
                    return returnError(messageContext, securityContext.getActingUser().get().getIdentifier() + " er ikke en " + requestedRole);
                } else {
                    log.warn(securityContext.getClient().get().getName() + " er ikke et " + requestedRole);
                    // Return requestedRole, because that is what the client sent to the backend
                    return returnError(messageContext, securityContext.getClient().get().getName() + " er ikke et " + requestedRole);
                }
            }
            RequestContext.get().setPermissions(authenticator.getPermissions());
            RequestContext.get().setRequestedRole(requestedRole);
            RequestContext.get().setValidatedRole(requestedRole);
            return true;
        } catch (DataAccessException e) {
            log.error("Error: ", e);
            return returnError(messageContext, "Intern server fejl");
        } catch (Exception e) {
            log.error("Error: ", e);
            return returnError(messageContext, e.getMessage());
        }
    }

    public OnBehalfOf getOnBehalfOf(DOMSource source, List<String> errorMessage) {
        OnBehalfOf result = new OnBehalfOf();

        NodeList onBehalfOf = source.getNode().getOwnerDocument().getElementsByTagNameNS("urn:sds-dk:ecpr2", "OnBehalfOf");
        if (onBehalfOf.getLength() > 0) {
            Node node = onBehalfOf.item(0);
            NodeList childNodes = node.getChildNodes();
            for (int i = 0; i < childNodes.getLength(); i++) {
                Node item = childNodes.item(i);
                if ("AuthorisationIdentifier".equals(item.getLocalName())) {
                    result.setOnBehalfOfAuthorisationNo(item.getTextContent());
                    break;
                }
                if ("PersonIdentifierType".equals(item.getLocalName())) {
                    NodeList personIdentifierChildren = item.getChildNodes();
                    for (int j = 0; j < personIdentifierChildren.getLength(); j++) {
                        Node personIdentifierChildItem = personIdentifierChildren.item(j);
                        if ("PersonIdentifier".equals(personIdentifierChildItem.getLocalName())) {
                            result.setOnBehalfOfPersonIdentifier(personIdentifierChildItem.getTextContent());
                            break;
                        }
                    }

                    break;
                    // attribute source=CPR is also in the request, we can extend SecurityCredential to use it if needed
                }
            }

        }
        return result;
    }

    @Override
    public boolean handleResponse(MessageContext messageContext, Object o) throws Exception {
        return true;
    }

    @Override
    public boolean handleFault(MessageContext messageContext, Object o) throws Exception {
        return true;
    }

    @Override
    public void afterCompletion(MessageContext messageContext, Object o, Exception e) throws Exception {

    }

    protected boolean returnError(MessageContext messageContext, String errorMessage) throws TransformerException {
        if (messageContext.getResponse() instanceof SoapMessage) {
            log.warn(errorMessage);
            SoapMessage response = (SoapMessage) messageContext.getResponse();
            SoapBody body = response.getSoapBody();
            SoapFault fault = body.addClientOrSenderFault("Validation error", Locale.ENGLISH);
            SoapFaultDetail detail = fault.addFaultDetail();

            SoapFaultDetailElement detailElement = detail.addFaultDetailElement(new QName("http://springframework.org/spring-ws", "ValidationError", "spring-ws"));
            detailElement.addText(errorMessage);
        }
        return false;
    }

    @Override
    public boolean shouldIntercept(MessageContext messageContext, Object o) {
        if (messageContext.getRequest() instanceof SoapMessage) {
            String soapAction = ((SoapMessage) messageContext.getRequest()).getSoapAction();
            if (StringUtils.hasLength(soapAction) && soapAction.charAt(0) == '"' && soapAction.charAt(soapAction.length() - 1) == '"') {
                soapAction = soapAction.substring(1, soapAction.length() - 1);
            }
            boolean nonInterceptingAction = NON_INTERCEPTING_SOAP_ACTIONS.contains(soapAction);
            if (nonInterceptingAction) {
                log.debug("Not validating role");
            }
            return !nonInterceptingAction;
        }
        return false;
    }
}
